#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 07:33
:Licence GNUv3
Part of grammpy

"""
import inspect
from ..Terminal import Terminal
from ..Nonterminal import Nonterminal
from ..HashContainer import HashContainer
from ..exceptions import NotNonterminalException, NotRuleException, TerminalDoesNotExistsException, \
    NonterminalDoesNotExistsException
from ..IsMethodsRuleExtension import IsMethodsRuleExtension


class RawGrammar:
    def __init__(self,
                 terminals=None,
                 nonterminals=None,
                 rules=None,
                 start_symbol=None):
        terminals = [] if terminals is None else terminals
        nonterminals = [] if nonterminals is None else nonterminals
        rules = [] if rules is None else rules
        self.__terminals = HashContainer()
        self.__nonterminals = HashContainer()
        self.__rules = HashContainer()
        self.__start_symbol = None
        self.add_term(terminals)
        self.add_nonterm(nonterminals)
        self.add_rule(rules)
        self.start_set(start_symbol)

    # Term part
    # TODO add validation of terminals that no rule or nonterminal is passed
    def add_term(self, term):
        return self.__terminals.add(term)

    def remove_term(self, term=None):
        return self.__terminals.remove(term)

    def have_term(self, term):
        return self.__terminals.have(term)

    def get_term(self, term=None):
        if term is not None and not HashContainer.is_iterable(term):
            item = self.__terminals.get(term)
            return Terminal(item, self) if item is not None else None
        vals = []
        obtain = self.__terminals.get(term)
        for t in obtain:
            vals.append(Terminal(t, self) if t is not None else None)
        return vals

    def term(self, term=None):
        return self.get_term(term)

    def terms(self):
        return [Terminal(term, self) for term in self.__terminals.all()]

    def terms_count(self):
        return self.__terminals.count()

    # Non term part
    @staticmethod
    def _controll_nonterms(nonterms):
        nonterms = HashContainer.to_iterable(nonterms)
        for nonterm in nonterms:
            if not inspect.isclass(nonterm) or not issubclass(nonterm, Nonterminal):
                raise NotNonterminalException(nonterm)
        return nonterms

    def add_nonterm(self, nonterms):
        nonterms = RawGrammar._controll_nonterms(nonterms)
        return self.__nonterminals.add(nonterms)

    def remove_nonterm(self, nonterms=None):
        if nonterms is None:
            return self.__nonterminals.remove()
        nonterms = RawGrammar._controll_nonterms(nonterms)
        return self.__nonterminals.remove(nonterms)

    def have_nonterm(self, nonterms):
        nonterms = RawGrammar._controll_nonterms(nonterms)
        return self.__nonterminals.have(nonterms)

    def get_nonterm(self, nonterms=None):
        if nonterms is None:
            return self.__nonterminals.get()
        converted = RawGrammar._controll_nonterms(nonterms)
        if not HashContainer.is_iterable(nonterms):
            return self.__nonterminals.get(converted)[0]
        return self.__nonterminals.get(converted)

    def nonterm(self, nonterms=None):
        return self.get_nonterm(nonterms)

    def nonterms(self):
        return self.__nonterminals.get()

    def nonterms_count(self):
        return self.__nonterminals.count()

    # Rules part
    def _control_rules(self, rules):
        rules = HashContainer.to_iterable(rules)
        for rule in rules:
            if not inspect.isclass(rule) or not issubclass(rule, IsMethodsRuleExtension):
                raise NotRuleException(rule)
            rule.validate(self)
        return rules

    def add_rule(self, rules):
        rules = self._control_rules(rules)
        return self.__rules.add(rules)

    def remove_rule(self, rules=None):
        if rules is None:
            return self.__rules.remove()
        rules = self._control_rules(rules)
        return self.__rules.remove(rules)

    def have_rule(self, rules):
        try:
            rules = self._control_rules(rules)
            return self.__rules.have(rules)
        except (TerminalDoesNotExistsException, NonterminalDoesNotExistsException):
            return False

    def get_rule(self, rules=None):
        if rules is None:
            return [rule for rule in self.__rules.get() if rule._active]
        converted = self._control_rules(rules)
        if not HashContainer.is_iterable(rules):
            return self.__rules.get(converted)[0]
        return self.__rules.get(converted)

    def rule(self, rules=None):
        return self.get_rule(rules)

    def rules(self):
        return self.rule()

    def rules_count(self):
        return len(self.rules())

    # StartSymbol
    def start_get(self):
        return self.__start_symbol

    def start_set(self, nonterminal):
        if nonterminal is None:
            self.__start_symbol = None
            return
        if not inspect.isclass(nonterminal) or not issubclass(nonterminal, Nonterminal):
            raise NotNonterminalException(nonterminal)
        if not self.have_nonterm(nonterminal):
            raise NonterminalDoesNotExistsException(None, nonterminal, self)
        self.__start_symbol = nonterminal

    def start_isSet(self):
        return self.__start_symbol is not None

    def start_is(self, nonterminal):
        return self.__start_symbol is nonterminal



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 08.07.2017 14:03
:Licence GNUv3
Part of grammpy

"""

from .GrammpyException import GrammpyException
from .NotNonterminalException import NotNonterminalException
from .NotASingleSymbolException import NotASingleSymbolException
from .RuleNotDefinedException import RuleNotDefinedException
from .CantCreateSingleRuleException import CantCreateSingleRuleException
from .CannotConvertException import CannotConvertException
from .RuleException import RuleException
from .RuleSyntaxException import RuleSyntaxException
from .UselessEpsilonException import UselessEpsilonException
from .TerminalDoesNotExistsException import TerminalDoesNotExistsException
from .NonterminalDoesNotExistsException import NonterminalDoesNotExistsException
from .NotRuleException import NotRuleException



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 08:54
:Licence GNUv3
Part of grammpy

"""

from .GrammpyException import GrammpyException

class CannotConvertException(GrammpyException):
    pass


#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 16:08
:Licence GNUv3
Part of grammpy

"""

from inspect import isclass
from .Rule import Rule
from .exceptions import RuleException, UselessEpsilonException, RuleSyntaxException, TerminalDoesNotExistsException, \
    NonterminalDoesNotExistsException
from .Constants import EPS
from .Nonterminal import Nonterminal


class IsMethodsRuleExtension(Rule):
    @staticmethod
    def _controlSide(cls, side, grammar):
        if not isinstance(side, list):
            raise RuleSyntaxException(cls, 'One side of rule is not enclose by list', side)
        if len(side) == 0:
            raise RuleSyntaxException(cls, 'One side of rule is not define', side)
        if EPS in side and len(side) > 1:
            raise UselessEpsilonException(cls)
        for symb in side:
            if isclass(symb) and issubclass(symb, Nonterminal):
                if not grammar.have_nonterm(symb):
                    raise NonterminalDoesNotExistsException(cls, symb, grammar)
            elif symb is EPS:
                continue
            elif not grammar.have_term(symb):
                raise TerminalDoesNotExistsException(cls, symb, grammar)

    @classmethod
    def validate(cls, grammar):
        r = cls.rules
        if not isinstance(r, list):
            raise RuleSyntaxException(cls, 'Rules property is not enclose in list')
        for rule in r:
            if not isinstance(rule, tuple):
                raise RuleSyntaxException(cls, 'One of the rules is not enclose in tuple', rule)
            if len(rule) != 2:
                raise RuleSyntaxException(cls, 'One of the rules does not have define left and right part', rule)
            l = rule[0]
            r = rule[1]
            IsMethodsRuleExtension._controlSide(cls, l, grammar)
            IsMethodsRuleExtension._controlSide(cls, r, grammar)
            if l == [EPS] and r == [EPS]:
                raise UselessEpsilonException(cls)

    @classmethod
    def is_valid(cls, grammar):
        try:
            cls.validate(grammar)
            return True
        except RuleException:
            return False



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 23.06.2017 16:45
:Licence GNUv3
Part of grammpy

"""

from .Grammars.PrettyApiGrammar import PrettyApiGrammar as Grammar
from .Nonterminal import Nonterminal
from .IsMethodsRuleExtension import IsMethodsRuleExtension as Rule
from .Constants import EPSILON, EPS



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 08:46
:Licence GNUv3
Part of grammpy

"""

from .RuleException import RuleException


class RuleNotDefinedException(RuleException):
    def __init__(self, rule):
        self.rule = rule



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 23.06.2017 16:45
:Licence GNUv3
Part of grammpy

"""

from .exceptions import CantCreateSingleRuleException, RuleNotDefinedException, NotASingleSymbolException


class CP(object):
    def __init__(self, getter):
        self._getter = getter

    def __get__(self, obj, cls=None):
        return self._getter(cls)  # for static remove cls from the call


class Rule:
    """
    fromSymbol = EPSILON
    toSymbol = EPSILON
    right = [EPSILON]
    left = [EPSILON]
    rule = ([EPSILON], [EPSILON])
    rules = [([EPSILON], [EPSILON])]
    """

    @CP
    def toSymbol(cls):
        if cls._traverse:
            raise RuleNotDefinedException(cls)
        if len(cls.rules) > 1:
            raise CantCreateSingleRuleException(cls)
        right = cls.rules[0][1]
        if len(right) > 1:
            raise NotASingleSymbolException(right)
        return right[0]

    @CP
    def fromSymbol(cls):
        if cls._traverse:
            raise RuleNotDefinedException(cls)
        if len(cls.rules) > 1:
            raise CantCreateSingleRuleException(cls)
        left = cls.rules[0][0]
        if len(left) > 1:
            raise NotASingleSymbolException(left)
        return left[0]

    @CP
    def right(cls):
        if cls._traverse:
            return [cls.toSymbol]
        if len(cls.rules) > 1:
            raise CantCreateSingleRuleException(cls)
        return cls.rules[0][1]

    @CP
    def left(cls):
        if cls._traverse:
            return [cls.fromSymbol]
        if len(cls.rules) > 1:
            raise CantCreateSingleRuleException(cls)
        return cls.rules[0][0]

    @CP
    def rule(cls):
        if cls._traverse:
            return (cls.left, cls.right)
        if len(cls.rules) > 1:
            raise CantCreateSingleRuleException(cls)
        return cls.rules[0]

    @CP
    def rules(cls):
        cls._traverse = True
        r = cls.rule
        cls._traverse = False
        return [r]

    _traverse = False

    _active = True

    @classmethod
    def rules_count(cls):
        return len(cls.rules)

    @classmethod
    def count(cls):
        return cls.rules_count()



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 04.07.2017 19:45
:Licence GNUv3
Part of grammpy

"""


class Terminal:
    '''
    Represent Terminal in former grammar
    '''

    def __init__(self, symbol, grammar):
        '''
        Create new terminal, where symbol is provided terminal passed into grammar
        :param symbol: Symbol representing terminal
        :param grammar: The grammar to which the terminal belongs
        '''
        self.__symbol = symbol
        self.__grammar = grammar

    def __hash__(self):
        '''
        Return hash based on symbol and instance of grammar
        :return: Hash value
        '''
        return hash((self.__symbol, id(self.__grammar)))

    def __eq__(self, other):
        '''
        Compare two terminals based on hash values
        :param other: Terminal to compare
        :type other: Terminal
        :return:
        :rtype bool
        '''
        return isinstance(other, Terminal) and hash(self) == hash(other)

    def symbol(self):
        '''
        Get symbol for the terminal.
        :return: Symbol that represent current terminal
        '''
        return self.__symbol

    @property
    def s(self):
        """
        Shortcut for symbol method
        :return: Symbol that represent current terminal
        """
        return self.symbol()



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 22.07.2017 13:19
:Licence GNUv3
Part of grammpy

"""
from collections import Iterable


class HashContainer:
    def __init__(self, items=None):
        if items is None:
            items = []
        self.__items = {}
        self.add(items)

    # Helpers
    @staticmethod
    def is_iterable(param: Iterable) -> bool:
        return isinstance(param, Iterable)

    @staticmethod
    def to_iterable(param: Iterable):
        # standardize it to iterable object
        if not HashContainer.is_iterable(param):
            return [param]
        return param

    def add(self, item):
        items = HashContainer.to_iterable(item)
        # iterace throught items
        for t in items:
            self.__items[hash(t)] = t

    def remove(self, item=None):
        if item is None:
            return self.__items.clear()
        item = HashContainer.to_iterable(item)
        # iterate throught items
        for t in item:
            del self.__items[hash(t)]

    def have(self, item):
        item = HashContainer.to_iterable(item)
        for t in item:
            if hash(t) not in self.__items:
                return False
        return True

    def get(self, item=None):
        # if no parameter is passed than return all terminals
        if item is None:
            # Maybe lazy evaluation, but it cannot be combined with return
            return self.all()
        # else return relevant to parameter
        transformed = HashContainer.to_iterable(item)
        ret = []
        for t in transformed:
            ret.append(t if hash(t) in self.__items else None)
        if not HashContainer.is_iterable(item):
            return ret[0]
        return ret

    def all(self):
        return [item for _, item in self.__items.items()]

    def count(self):
        return len(self.__items)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 04.08.2017 19:00
:Licence GNUv3
Part of grammpy

"""

from distutils.core import setup


setup(
    name='grammpy',
    version='1.1.0',
    packages=['grammpy', 'grammpy.Grammars', 'grammpy.exceptions'],
    url='https://github.com/PatrikValkovic/grammpy',
    license='GNU General Public License v3.0',
    author='Patrik Valkovic',
    download_url='https://github.com/PatrikValkovic/grammpy/archive/v1.0.1.tar.gz',
    author_email='patrik.valkovic@hotmail.cz',
    description='Package for representing formal grammars.'
)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 07:33
:Licence GNUv3
Part of grammpy

"""

from grammpy.Grammars.RawGrammar import RawGrammar as Grammar


class StringGrammar(Grammar):
    @staticmethod
    def __to_string_arr(t):
        if isinstance(t, str):
            return [t]
        return t

    def remove_term(self, term=None):
        return super().remove_term(StringGrammar.__to_string_arr(term))

    def add_term(self, term):
        return super().add_term(StringGrammar.__to_string_arr(term))

    def term(self, term=None):
        return self.get_term(term)

    def get_term(self, term=None):
        res = super().get_term(StringGrammar.__to_string_arr(term))
        if isinstance(term, str):
            return res[0]
        return res

    def have_term(self, term):
        return super().have_term(StringGrammar.__to_string_arr(term))



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 01.08.2017 07:33
:Licence GNUv3
Part of grammpy

"""

from grammpy.Grammars.StringGrammar import StringGrammar as Grammar


class PrettyApiGrammar(Grammar):
    def __init__(self,
                 terminals=None,
                 nonterminals=None,
                 rules=None,
                 start_symbol=None):
        if isinstance(terminals, str):
            temp = []
            for ch in terminals:
                temp.append(ch)
            terminals = temp
        super().__init__(terminals=terminals,
                         nonterminals=nonterminals,
                         rules=rules,
                         start_symbol=start_symbol)
